@@ -1,5 +1,6 @@
 
 import io
+import ssl
 import copy
 import typing
 import asyncio
@@ -58,6 +59,7 @@ from aardwolf.protocol.T128.fontlistpdu import TS_FONT_LIST_PDU
 from aardwolf.protocol.T128.inputeventpdu import TS_SHAREDATAHEADER, TS_INPUT_EVENT, TS_INPUT_PDU_DATA
 from aardwolf.protocol.T125.securityexchangepdu import TS_SECURITY_PACKET
 from aardwolf.protocol.T128.seterrorinfopdu import TS_SET_ERROR_INFO_PDU
+from aardwolf.protocol.T128.shutdownreqpdu import TS_SHUTDOWN_REQ_PDU
 from aardwolf.protocol.T128.share import PDUTYPE, STREAM_TYPE, PDUTYPE2
 
 
@@ -124,6 +126,7 @@ class RDPConnection:
 		self.cryptolayer:RDPCryptoLayer = None
 		self.__desktop_buffer = None
 		self.desktop_buffer_has_data = False
+		self.__terminate_called = False
 
 		self.__vk_to_sc = {
 			'VK_BACK'     : 14,
@@ -176,14 +179,25 @@ class RDPConnection:
 
 	
 	async def terminate(self):
+		"""sends a shutdown request to the server and terminates the connection"""
 		try:
+			if self.__terminate_called is True:
+				return True, None
+			self.__terminate_called = True
+
+			will_shutdown, err = await self.send_disconnect()
+			if err is not None:
+				logger.warning('Error while requesting shutdown')
+			else:
+				if will_shutdown is False:
+					logger.warning('Server refused to shutdown, proceeding with termination anyway...')
+
 			for name in self.__joined_channels:
 				await self.__joined_channels[name].disconnect()
 			
 			if self.ext_out_queue is not None:
 				# signaling termination via ext_out_queue
-				await self.ext_out_queue.put(None)
-				
+				await self.ext_out_queue.put(None)			
 			
 			if self.__external_reader_task is not None:
 				self.__external_reader_task.cancel()
@@ -193,10 +207,12 @@ class RDPConnection:
 			
 			return True, None
 		except Exception as e:
-			traceback.print_exc()
+			logger.error(f"Error: {e}, {traceback.format_exc()}")
 			return None, e
 		finally:
 			self.disconnected_evt.set()
+			if self.__connection is not None:
+				await self.__connection.close()
 	
 	async def __aenter__(self):
 		return self
@@ -256,7 +272,14 @@ class RDPConnection:
 				self.x224_flag = self.x224_connection_reply.flags
 				logger.debug('Server selected protocol: %s' % self.x224_protocol)
 				if SUPP_PROTOCOLS.SSL in self.x224_protocol or SUPP_PROTOCOLS.HYBRID in self.x224_protocol or SUPP_PROTOCOLS.HYBRID_EX in self.x224_protocol:
-					await self.__connection.wrap_ssl()
+					if self.target.unsafe_ssl:
+						ssl_ctx = ssl.create_default_context()
+						ssl_ctx.check_hostname = False
+						ssl_ctx.verify_mode = ssl.CERT_NONE
+						ssl_ctx.set_ciphers('ALL:@SECLEVEL=0')
+						await self.__connection.wrap_ssl(ssl_ctx=ssl_ctx)
+					else:
+						await self.__connection.wrap_ssl()
 
 				# if the server expects HYBRID/HYBRID_EX authentication we do that here
 				# This is basically credSSP
@@ -547,7 +570,9 @@ class RDPConnection:
 			conf_create_req = self.__t125_ber_codec.encode('ConnectMCSPDU',('connect-initial', initialconnect))
 			await self._x224net.write(bytes(conf_create_req))
 			
-			response_raw = await self._x224net.read()			
+			response_raw = await self._x224net.read()
+			if response_raw is None:
+				raise Exception('Connection closed!')
 			server_res_raw = response_raw.data
 			server_res_t125 = self.__t125_ber_codec.decode('ConnectMCSPDU', server_res_raw)
 			if server_res_t125[0] != 'connect-response':
@@ -586,7 +611,7 @@ class RDPConnection:
 
 			return True, None
 		except Exception as e:
-			traceback.print_exc()
+			logger.error(f"Error: {e}, {traceback.format_exc()}")
 			return None, e
 
 	async def __erect_domain(self):
@@ -604,6 +629,8 @@ class RDPConnection:
 			request = self._t125_per_codec.encode('DomainMCSPDU', ('attachUserRequest', {}))
 			await self._x224net.write(request)
 			response = await self._x224net.read()
+			if response is None:
+				raise Exception('Connection closed!')
 			response_parsed = self._t125_per_codec.decode('DomainMCSPDU', response.data)
 			if response_parsed[0] != 'attachUserConfirm':
 				raise Exception('Unexpected response! %s' % response_parsed[0])
@@ -621,6 +648,8 @@ class RDPConnection:
 				joindata = self._t125_per_codec.encode('DomainMCSPDU', ('channelJoinRequest', {'initiator': self._initiator, 'channelId': self.__joined_channels[name].channel_id}))
 				await self._x224net.write(bytes(joindata))
 				response = await self._x224net.read()
+				if response is None:
+					raise Exception('Connection closed!')
 				
 				x = self._t125_per_codec.decode('DomainMCSPDU', response.data)
 				if x[0] != 'channelJoinConfirm':
@@ -725,7 +754,7 @@ class RDPConnection:
 
 			return True, None
 		except Exception as e:
-			traceback.print_exc()
+			logger.error(f"Error: {e}, {traceback.format_exc()}")
 			return None, e
 	
 	async def __handle_mandatory_capability_exchange(self):
@@ -917,6 +946,64 @@ class RDPConnection:
 			return True, None
 		except Exception as e:
 			return None, e
+		
+	async def send_disconnect(self):
+		"""Sends a disconnect request to the server. This will NOT close the connection!"""
+		try:
+			data_start_offset = 0
+			if self.__server_connect_pdu[TS_UD_TYPE.SC_SECURITY].encryptionLevel == 1:
+				# encryptionLevel == 1 means that server data is not encrypted. This results in this part of the negotiation 
+				# that the server sends data to the client with an empty security header (which is not documented....)
+				data_start_offset = 4
+			
+			data_hdr = TS_SHAREDATAHEADER()
+			data_hdr.shareID = 0x103EA
+			data_hdr.streamID = STREAM_TYPE.MED
+			data_hdr.pduType2 = PDUTYPE2.SHUTDOWN_REQUEST
+			
+			
+			cli_input = TS_INPUT_PDU_DATA()
+			cli_input.slowPathInputEvents.append(TS_SHUTDOWN_REQ_PDU())
+
+			sec_hdr = None
+			if self.cryptolayer is not None:
+				sec_hdr = TS_SECURITY_HEADER()
+				sec_hdr.flags = SEC_HDR_FLAG.ENCRYPT
+				sec_hdr.flagsHi = 0
+
+			await self.handle_out_data(cli_input, sec_hdr, data_hdr, None, self.__joined_channels['MCS'].channel_id, False)
+			data, err = await self.__joined_channels['MCS'].out_queue.get()
+			if err is not None:
+				raise err
+			
+			server_shutdown_reply = False
+			data = data[data_start_offset:]
+			shc = TS_SHARECONTROLHEADER.from_bytes(data)
+			if shc.pduType == PDUTYPE.DATAPDU:
+				shd = TS_SHAREDATAHEADER.from_bytes(data)
+				if shd.pduType2 == PDUTYPE2.SHUTDOWN_DENIED:
+					# server denied ur request
+					server_shutdown_reply = False
+				elif shd.pduType2 == PDUTYPE2.CONTROL:
+					res_control = TS_CONTROL_PDU.from_bytes(data)
+					if res_control.action == CTRLACTION.COOPERATE:
+						# server will cooperate
+						server_shutdown_reply = True
+					else:
+						# I dunno what the server is thinking
+						server_shutdown_reply = False
+
+				else:
+					# I dunno what the server is thinking
+					# Maybe we consumed the wrong packet?
+					server_shutdown_reply = False
+			else:
+				raise Exception('Unexpected reply! %s' % shc.pduType.name)
+
+			
+			return server_shutdown_reply, None
+		except Exception as e:
+			return None, e
 
 	async def __x224_reader(self):
 		# recieves X224 packets and fastpath packets, performs decryption if necessary then dispatches each packet to 
@@ -982,7 +1069,7 @@ class RDPConnection:
 		except asyncio.CancelledError:
 			return None, None
 		except Exception as e:
-			traceback.print_exc()
+			logger.error(f"Error: {e}, {traceback.format_exc()}")
 			return None, e
 		finally:
 			await self.terminate()
@@ -1013,7 +1100,7 @@ class RDPConnection:
 			#	#	print('notbitmap %s' % fpdu.fpOutputUpdates.updateCode.name)
 		except Exception as e:
 			# the decoder is not perfect yet, so it's better to keep this here...
-			traceback.print_exc()
+			logger.error(f"Error: {e}, {traceback.format_exc()}")
 			return
 	
 
@@ -1027,7 +1114,7 @@ class RDPConnection:
 				scancode = scancode_hint
 			return await self.send_key_scancode(scancode, is_pressed, is_extended)
 		except Exception as e:
-			traceback.print_exc()
+			logger.error(f"Error: {e}, {traceback.format_exc()}")
 			return None, e
 	
 	async def send_key_scancode(self, scancode, is_pressed, is_extended, modifiers = VK_MODIFIERS(0)):
@@ -1058,7 +1145,7 @@ class RDPConnection:
 				
 
 		except Exception as e:
-			traceback.print_exc()
+			logger.error(f"Error: {e}, {traceback.format_exc()}")
 			return None, e
 
 	async def send_key_char(self, char, is_pressed):
@@ -1087,10 +1174,10 @@ class RDPConnection:
 			return True, None
 
 		except Exception as e:
-			traceback.print_exc()
+			logger.error(f"Error: {e}, {traceback.format_exc()}")
 			return None, e
 
-	async def send_mouse(self, button:MOUSEBUTTON, xPos:int, yPos:int, is_pressed:bool):
+	async def send_mouse(self, button:MOUSEBUTTON, xPos:int, yPos:int, is_pressed:bool, steps:int = 0):
 		try:
 			if xPos < 0 or yPos < 0:
 				return True, None
@@ -1113,6 +1200,14 @@ class RDPConnection:
 				# indicates a simple pointer update with no buttons pressed
 				# sending this enables the mouse hover feel on the remote end
 				mouse.pointerFlags |= PTRFLAGS.MOVE
+			if button == MOUSEBUTTON.MOUSEBUTTON_WHEEL_UP:
+				mouse.pointerFlags |= PTRFLAGS.WHEEL
+				mouse.pointerFlags |= (PTRFLAGS.WheelRotationMask & steps)
+
+			if button == MOUSEBUTTON.MOUSEBUTTON_WHEEL_DOWN:
+				mouse.pointerFlags |= PTRFLAGS.WHEEL_NEGATIVE
+				mouse.pointerFlags |= (PTRFLAGS.WheelRotationMask & steps)
+
 			mouse.xPos = xPos
 			mouse.yPos = yPos
 
@@ -1130,7 +1225,7 @@ class RDPConnection:
 					
 			await self.handle_out_data(cli_input, sec_hdr, data_hdr, None, self.__joined_channels['MCS'].channel_id, False)
 		except Exception as e:
-			traceback.print_exc()
+			logger.error(f"Error: {e}, {traceback.format_exc()}")
 			return None, e
 
 	def get_desktop_buffer(self, encoding:VIDEO_FORMAT = VIDEO_FORMAT.PIL):
@@ -1148,7 +1243,7 @@ class RDPConnection:
 			else:
 				raise ValueError('Output format of "%s" is not supported!' % encoding)
 		except Exception as e:
-			traceback.print_exc()
+			logger.error(f"Error: {e}, {traceback.format_exc()}")
 			return None, e
 	
 	async def get_current_clipboard_text(self):
@@ -1214,7 +1309,7 @@ class RDPConnection:
 			return None, None
 
 		except Exception as e:
-			traceback.print_exc()
+			logger.error(f"Error: {e}, {traceback.format_exc()}")
 			await self.terminate()
 			return None, e
 	
@@ -1273,7 +1368,7 @@ class RDPConnection:
 				raise NotImplementedError("Fastpath output is not yet implemented")
 
 		except Exception as e:
-			traceback.print_exc()
+			logger.error(f"Error: {e}, {traceback.format_exc()}")
 			await self.terminate()
 			return None, e
 		
@@ -1285,7 +1380,7 @@ async def amain():
 		from aardwolf.extensions.RDPEDYC.channel import RDPEDYCChannel
 
 		iosettings = RDPIOSettings()
-		url = 'rdp+ntlm-password://TEST\\Administrator:Passw0rd!1@10.10.10.102'
+		url = 'rdp+ntlm-password://TEST2\\Administrator:Passw0rd!1@192.168.85.131'
 		rdpurl = RDPConnectionFactory.from_url(url, iosettings)
 		conn = rdpurl.get_connection(iosettings)
 		_, err = await conn.connect()
